package com.dbxml.db.client.xmldb;

/*
 * dbXML - Native XML Database
 * Copyright (C) 1999-2004  The dbXML Group, L.L.C.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Id: XSLTransformService.java,v 1.2 2004/02/12 00:17:56 bradford Exp $
 */

import org.xmldb.api.base.Resource;
import org.xmldb.api.base.Service;
import org.xmldb.api.base.XMLDBException;
import org.xmldb.api.modules.XMLResource;

/**
 * XSLTransformService is a <code>Service</code> that enables the execution of
 * XSL transformations within the context of a <code>Collection</code> or
 * against a single XML <code>Resource</code> stored in the
 * <code>Collection</code>.
 */

public interface XSLTransformService extends Service {
   // Namespace Manipulation

   /**
    * Sets a namespace mapping in the internal namespace map used to evaluate
    * queries. If <code>prefix</code> is null or empty the default namespace is
    * associated with the provided URI. A null or empty <code>uri</code> results
    * in an exception being thrown.
    *
    * @param prefix The prefix to set in the map. If
    *  <code>prefix</code> is empty or null the
    *  default namespace will be associated with the provided URI.
    * @param uri The URI for the namespace to be associated with prefix.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    * @todo probably need some special error here.
    */
   void setNamespace(String prefix, String uri) throws XMLDBException;

   /**
    * Returns the URI string associated with <code>prefix</code> from
    * the internal namespace map. If <code>prefix</code> is null or empty the
    * URI for the default namespace will be returned. If a mapping for the
    * <code>prefix</code> can not be found null is returned.
    *
    * @param prefix The prefix to retrieve from the namespace map.
    * @return The URI associated with <code>prefix</code>
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    */
   String getNamespace(String prefix) throws XMLDBException;

   /**
    * Removes the namespace mapping associated with <code>prefix</code> from
    * the internal namespace map. If <code>prefix</code> is null or empty the
    * mapping for the default namespace will be removed.
    *
    * @param prefix The prefix to remove from the namespace map. If
    *  <code>prefix</code> is null or empty the mapping for the default
    *  namespace will be removed.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    */
   void removeNamespace(String prefix) throws XMLDBException;

   /**
    * Removes all namespace mappings stored in the internal namespace map.
    *
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    */
   void clearNamespaces() throws XMLDBException;


   // Templates

   /**
    * Sets the templates that will be used to transform this document.   The
    * templates are retrieved from the XSL stylesheet document that the
    * templates XMLResource represents.
    *
    * @param templates The XSL stylesheet that will be used for the
    *  transformation.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    */
   void setTemplates(XMLResource templates) throws XMLDBException;


   // XSLT Bound Parameters

   /**
    * Sets a bound parameter.  The value of this parameter will be passed
    * into the XSL processor and bound to the variable name for processing.
    * XSL parameters are declared with the param element.
    *
    * @param name The bound parameter name
    * @param value The bound parameter value
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    */
   void setParameter(String name, String value) throws XMLDBException;

   /**
    * Retrieves the value of a bound parameter, or null if the parameter
    * has not already been bound.
    *
    * @param name The bound parameter value to retrieve
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    */
   String getParameter(String name) throws XMLDBException;

   /**
    * Removes or un-binds a bound parameter.
    *
    * @param name The bound paramter name to remove.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    */
   void removeParameter(String name) throws XMLDBException;

   /**
    * Removes or un-binds *all* bound paramters.
    *
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    */
   void clearParameters() throws XMLDBException;


   // Query Invocation

   /**
    * Uses the defined templates to trasform the results of the XPath query
    * against the <code>Collection</code>. The XPath will be applied to all
    * XML resources stored in the <code>Collection</code>.  The result is a
    * <code>ResourceSet</code> containing the results of the query. Any
    * namespaces used in the <code>query</code> string will be evaluated using
    * the mappings setup using <code>setNamespace</code>.
    *
    * @param query The XPath query string to use.
    * @return A <code>ResourceSet</code> containing the results of the query.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    */
   Resource transform(String xpath) throws XMLDBException;

   /**
    * Uses the defined templates to transform the XML resource stored in the
    * <code>Collection</code> associated with this service. The result is a
    * <code>ResourceSet</code> containing the results of the query. Any
    * namespaces used in the <code>query</code> string will be evaluated using
    * the mappings setup using <code>setNamespace</code>.
    *
    * @param id The id of the document to run the query against.
    * @return A <code>ResourceSet</code> containing the results of the query.
    * @exception XMLDBException with expected error codes.<br />
    *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
    *  specific errors that occur.<br />
    */
   Resource transformResource(String id) throws XMLDBException;
}
